// R3DModel.cpp: implementation of the R3DModel class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
//#include "Traffic3D.h"
#include "R3DModel.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// Notes:
// An R3DModel is made up of an array of vectors which are rotated. moved etc.
// Prior to drawing, the squares, triangles, cubes and cylinders are mapped onto the appropriate
// vectors.  This increases memory requirements but decreases the number of rotations and translations
// that are necessary, as most vectors are shared by a number of shapes.

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

R3DModel::R3DModel()
{
	mNumVectors=0;
	mNumTriangles=0;
	mNumRectangles=0;
	mNumCylinders=0;
	mNumCuboids=0;
}

R3DModel::~R3DModel()
{
	mVectors.RemoveAll();
	mTriangles.RemoveAll();
	mRectangles.RemoveAll();
	mCuboids.RemoveAll();
	mCylinders.RemoveAll();
}

// This constructor initialises the arrays to a set size to reduce memory fragmentation
R3DModel::R3DModel(int ntriangles, int nrectangles, int ncuboids, int ncylinders)
{
	// does this have to be repeated here?  Or is the default constructor always called?
	mNumVectors=0;
	mNumTriangles=0;
	mNumRectangles=0;
	mNumCylinders=0;
	mNumCuboids=0;

	mTriangles.SetSize(ntriangles);
	mTriangleMap.SetSize(ntriangles*3);	// 3 vectors/triangle
	mRectangles.SetSize(nrectangles);
	mRectangleMap.SetSize(nrectangles*4);	// 4 vectors/rectangle
	mCuboids.SetSize(ncuboids);
	mCylinders.SetSize(ncylinders);
}

// adds vectors to the array.  A pointer to the 1st one and the number of vectors to add are specified
void R3DModel::SetupVectors(R3DVector *v, int number)
	{
	mNumVectors+=number;	// update the total number of vectors in the array
	mVectors.SetSize(mNumVectors);	// resize the array appropriately

	for(int x=0; x<number; x++)	// for each new vector
		{
		mVectors.SetAtGrow(x, *v);	// is the dereference necessary or is pointer->reference conversion auto?
		v++;	// increment the pointer to point to next vector
		}
	}

// adds a triangle to the model by specifying the triangle and the vectors it is mapped on to
void R3DModel::AddTriangle(R3DTriangle &t, int v1, int v2, int v3)
{
	if(v1>=mNumVectors || v2>=mNumVectors || v3>=mNumVectors)	// if any are out of range
		return;	// replace this with proper error handling later
	
	int i=mNumTriangles*3;	// 3 vectors per triangle
	mTriangles.SetAtGrow(mNumTriangles, t);	// add the triangles
	mTriangleMap.SetAtGrow(i, v1);
	mTriangleMap.SetAtGrow(i+1, v2);
	mTriangleMap.SetAtGrow(i+2, v3);	// update the map
	mNumTriangles++;
}


// same as the triangle add function, but with rectangles!
void R3DModel::AddRectangle(R3DRectangle &r, int v1, int v2, int v3, int v4)
{
	if(v1>=mNumVectors || v2>=mNumVectors || v3>=mNumVectors || v4>=mNumVectors)	// if any are out of range
		return;	// replace this with proper error handling later
	
	int i=mNumRectangles*4;	// 4 vectors/rectangle
	mRectangles.SetAtGrow(mNumRectangles, r);	// add the rectangles
	mRectangleMap.SetAtGrow(i, v1);
	mRectangleMap.SetAtGrow(i+1, v2);
	mRectangleMap.SetAtGrow(i+2, v3);
	mRectangleMap.SetAtGrow(i+3, v4);	// update the map
	mNumRectangles++;
}

// adds a cuboid to the array.  Maps aren't used because cuboids
// handle that themselves
void R3DModel::AddCuboid(R3DCuboid &c)
{
	mCuboids.SetAtGrow(mNumCuboids, c);
	mNumCuboids++;
}

// same as for cuboids
void R3DModel::AddCylinder(R3DCylinder &c)
{
	mCylinders.SetAtGrow(mNumCylinders, c);
	mNumCylinders++;
}

// rotates all the vertexes, cuboids and cylinders
void R3DModel::RotateX(R3DVector Origin, float Theta)
{
	// vertexes:
	for(int x=1; x<=mNumVectors; x++)
		{
		R3DVector v=mVectors.GetAt(x-1);
		v.RotateX(Origin, Theta);
		mVectors.SetAt(x-1, v);
		}

	// cuboids:
	for(x=1; x<=mNumCuboids; x++)
		{
		R3DCuboid c=mCuboids.GetAt(x-1);
		c.RotateX(Origin, Theta);
		mCuboids.SetAt(x-1, c);
		}

	// cylinders:
	for(x=1; x<=mNumCylinders; x++)
		{
		R3DCylinder y=mCylinders.GetAt(x-1);
		y.RotateX(Origin, Theta);
		mCylinders.SetAt(x-1, y);
		}
}

// as RotateX
void R3DModel::RotateY(R3DVector Origin, float Theta)
{
	// vertexes:
	for(int x=1; x<=mNumVectors; x++)
		{
		R3DVector v=mVectors.GetAt(x-1);
		v.RotateY(Origin, Theta);
		mVectors.SetAt(x-1, v);
		}

	// cuboids:
	for(x=1; x<=mNumCuboids; x++)
		{
		R3DCuboid c=mCuboids.GetAt(x-1);
		c.RotateY(Origin, Theta);
		mCuboids.SetAt(x-1, c);
		}

	// cylinders:
	for(x=1; x<=mNumCylinders; x++)
		{
		R3DCylinder y=mCylinders.GetAt(x-1);
		y.RotateY(Origin, Theta);
		mCylinders.SetAt(x-1, y);
		}
}

// as RotateX and RotateY:
void R3DModel::RotateZ(R3DVector Origin, float Theta)
{
	// vertexes:
	for(int x=1; x<=mNumVectors; x++)
		{
		R3DVector v=mVectors.GetAt(x-1);
		v.RotateZ(Origin, Theta);
		mVectors.SetAt(x-1, v);
		}

	// cuboids:
	for(x=1; x<=mNumCuboids; x++)
		{
		R3DCuboid c=mCuboids.GetAt(x-1);
		c.RotateZ(Origin, Theta);
		mCuboids.SetAt(x-1, c);
		}

	// cylinders:
	for(x=1; x<=mNumCylinders; x++)
		{
		R3DCylinder y=mCylinders.GetAt(x-1);
		y.RotateZ(Origin, Theta);
		mCylinders.SetAt(x-1, y);
		}
}

// does the same job as the rotates, but moves instead
void R3DModel::Move(R3DVector v)
{
		// vertexes:
	for(int x=1; x<=mNumVectors; x++)
		{
		R3DVector Oldv=mVectors.GetAt(x-1);
		Oldv+=v;
		mVectors.SetAt(x-1, Oldv);
		}

	// cuboids:
	for(x=1; x<=mNumCuboids; x++)
		{
		R3DCuboid c=mCuboids.GetAt(x-1);
		c.Move(v);
		mCuboids.SetAt(x-1, c);
		}

	// cylinders:
	for(x=1; x<=mNumCylinders; x++)
		{
		R3DCylinder y=mCylinders.GetAt(x-1);
		y.Move(v);
		mCylinders.SetAt(x-1, y);
		}
}

// draws the model by assigning positioning the triangles and rectangles according to their maps. 
void R3DModel::Draw()
{
	int i=0;	// used for the map indexes
	R3DVector v1, v2, v3, v4;	// used to map triangles & rectangles
	for(int x=1; x<=mNumTriangles; x++)
		{
		R3DTriangle t=mTriangles.GetAt(x-1);
		i=(x-1)*3;	// the first map index for the triangle
		v1=mVectors.GetAt(mTriangleMap.GetAt(i));	// get the first vector mapped by the triangle
		v2=mVectors.GetAt(mTriangleMap.GetAt(i+1));
		v3=mVectors.GetAt(mTriangleMap.GetAt(i+2));
		t.SetPosition(v1, v2, v3);
		t.Draw();	// actually draw the darn thing!
		}

	// now do the rectangles in the same way
	for(x=1; x<=mNumRectangles; x++)
		{
		R3DRectangle r=mRectangles.GetAt(x-1);
		i=(x-1)*4;	// there are 4 vectors for each rectangle
		v1=mVectors.GetAt(mRectangleMap.GetAt(i));
		v2=mVectors.GetAt(mRectangleMap.GetAt(i+1));
		v3=mVectors.GetAt(mRectangleMap.GetAt(i+2));
		v4=mVectors.GetAt(mRectangleMap.GetAt(i+3));
		r.SetPosition(v1, v2, v3, v4);
		r.Draw();
		}

	// now simply draw the cuboids:
	for(x=1; x<=mNumCuboids; x++)
		{
		R3DCuboid c=mCuboids.GetAt(x-1);
		c.Draw();
		}

	// ...and the cylinders too
	for(x=1; x<=mNumCylinders; x++)
		{
		R3DCylinder y=mCylinders.GetAt(x-1);
		y.Draw();
		}
}

void R3DModel::SetTexture(R3DTexture t)
{
	// No action for now
}

// models can be serialised:
IMPLEMENT_SERIAL( R3DModel, CObject, 1 )

void R3DModel::Serialize(CArchive &archive)
{
	CObject::Serialize(archive);	// call base class serialize function

	if(archive.IsStoring())
		archive << mNumVectors << mNumTriangles << mNumRectangles << mNumCuboids << mNumCylinders;
	else
		archive >> mNumVectors >> mNumTriangles >> mNumRectangles >> mNumCuboids >> mNumCylinders;

	mVectors.Serialize(archive);
	mTriangles.Serialize(archive);
	mTriangleMap.Serialize(archive);
	mRectangles.Serialize(archive);
	mRectangleMap.Serialize(archive);
	mCuboids.Serialize(archive);
	mCylinders.Serialize(archive);
}

// The SerializeElements functions for triangles, rectangles, cuboids and cylinders:
// Triangles:
template <> void AFXAPI SerializeElements <R3DTriangle> ( CArchive& ar, R3DTriangle* t, int nCount )
	{
	for(int x=0; x<nCount; x++, t++)
		{
		// serialize each object in the array:
		t->Serialize(ar);
		}
	}

// Rectangles:
template <> void AFXAPI SerializeElements <R3DRectangle> ( CArchive& ar, R3DRectangle* t, int nCount )
	{
	for(int x=0; x<nCount; x++, t++)
		{
		// serialize each object in the array:
		t->Serialize(ar);
		}
	}

// Cuboids:
template <> void AFXAPI SerializeElements <R3DCuboid> ( CArchive& ar, R3DCuboid* t, int nCount )
	{
	for(int x=0; x<nCount; x++, t++)
		{
		// serialize each object in the array:
		t->Serialize(ar);
		}
	}

// cylinders:
template <> void AFXAPI SerializeElements <R3DCylinder> ( CArchive& ar, R3DCylinder* t, int nCount )
	{
	for(int x=0; x<nCount; x++, t++)
		{
		// serialize each object in the array:
		t->Serialize(ar);
		}
	}

template <> void AFXAPI SerializeElements <int> ( CArchive& ar, int* t, int nCount )
	{
	if(ar.IsStoring())
		{
		for(int x=0; x<nCount; x++, t++)
			ar << *t;
		}
	else
		{
		for(int x=0; x<nCount; x++, t++)
			ar >> *t;
		}
	}
